@@ -3,6 +3,7 @@ import datetime
 import pytest
 
 from django import test
+from django.contrib.auth.models import Permission
 from django.test.utils import override_settings
 from django.core.exceptions import ImproperlyConfigured, PermissionDenied
 from django.http import Http404, HttpResponse
@@ -11,7 +12,12 @@ from django.utils.timezone import make_aware, get_current_timezone
 
 from django.urls import reverse_lazy
 
-from .factories import GroupFactory, UserFactory
+from .factories import (
+    GroupFactory,
+    UserFactory,
+    UserObjectPermissionsFactory,
+    ArticleFactory,
+)
 from .helpers import TestViewHelper
 from .views import (
     PermissionRequiredView,
@@ -67,9 +73,7 @@ class _TestAccessBasicsMixin(TestViewHelper):
         user = self.build_unauthorized_user()
         self.client.login(username=user.username, password="asdf1234")
         resp = self.client.get(self.view_url)
-        self.assertRedirects(
-            resp, "/accounts/login/?next={0}".format(self.view_url)
-        )
+        self.assertRedirects(resp, "/accounts/login/?next={0}".format(self.view_url))
 
     def test_raise_permission_denied(self):
         """
@@ -158,15 +162,11 @@ class _TestAccessBasicsMixin(TestViewHelper):
         user = self.build_unauthorized_user()
         req = self.build_request(user=user, path=self.view_url)
         resp = self.dispatch_view(req, login_url="/login/")
-        self.assertEqual(
-            "/login/?next={0}".format(self.view_url), resp["Location"]
-        )
+        self.assertEqual("/login/?next={0}".format(self.view_url), resp["Location"])
 
         # Test with reverse_lazy
         resp = self.dispatch_view(req, login_url=reverse_lazy("headline"))
-        self.assertEqual(
-            "/headline/?next={0}".format(self.view_url), resp["Location"]
-        )
+        self.assertEqual("/headline/?next={0}".format(self.view_url), resp["Location"])
 
     def test_custom_redirect_field_name(self):
         """
@@ -185,9 +185,7 @@ class _TestAccessBasicsMixin(TestViewHelper):
         ImproperlyConfigured.
         """
         with self.assertRaises(ImproperlyConfigured):
-            self.dispatch_view(
-                self.build_request(path=self.view_url), login_url=None
-            )
+            self.dispatch_view(self.build_request(path=self.view_url), login_url=None)
 
     def test_get_redirect_field_name_raises_exception(self):
         """
@@ -209,9 +207,7 @@ class _TestAccessBasicsMixin(TestViewHelper):
         user = self.build_unauthorized_user()
         self.client.login(username=user.username, password="asdf1234")
         resp = self.client.get(self.view_url)
-        self.assertRedirects(
-            resp, "/auth/login/?next={0}".format(self.view_url)
-        )
+        self.assertRedirects(resp, "/auth/login/?next={0}".format(self.view_url))
 
     def test_redirect_unauthenticated(self):
         resp = self.dispatch_view(
@@ -220,9 +216,7 @@ class _TestAccessBasicsMixin(TestViewHelper):
             redirect_unauthenticated_users=True,
         )
         assert resp.status_code == 302
-        assert resp["Location"] == "/accounts/login/?next={0}".format(
-            self.view_url
-        )
+        assert resp["Location"] == "/accounts/login/?next={0}".format(self.view_url)
 
     def test_redirect_unauthenticated_false(self):
         with self.assertRaises(PermissionDenied):
@@ -413,12 +407,55 @@ class TestPermissionRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
         with self.assertRaises(ImproperlyConfigured):
             self.dispatch_view(self.build_request(), permission_required=None)
 
+    def test_object_level_permissions(self):
+        """
+        Tests that object level permissions perform as expected, where object level permissions and
+        global level permissions
+        """
+        # Arrange
+        article = ArticleFactory()
+        self.view_class = PermissionRequiredView
+        self.view_url = f"/object_level_permission_required/?pk={article.pk}"
+        tests_add_article = Permission.objects.get(codename="add_article")
+        permissions = "tests.add_article"
+        valid_user = UserFactory(permissions=[permissions])
+        invalid_user_1 = UserFactory(permissions=["auth.add_user"])
+        invalid_user_2 = UserFactory(permissions=[permissions])
+        UserObjectPermissionsFactory(
+            user=valid_user, permission=tests_add_article, article_object=article
+        )
+        # Act
+        valid_req = self.build_request(path=self.view_url, user=valid_user)
+        valid_resp = self.dispatch_view(
+            valid_req,
+            permission_required=permissions,
+            object_level_permissions=True,
+            raise_exception=True,
+        )
+        invalid_req_1 = self.build_request(path=self.view_url, user=invalid_user_1)
+        invalid_req_2 = self.build_request(path=self.view_url, user=invalid_user_2)
+        # Assert
+        self.assertEqual(valid_resp.status_code, 200)
+        with self.assertRaises(PermissionDenied):
+            self.dispatch_view(
+                invalid_req_1,
+                permission_required=permissions,
+                object_level_permissions=True,
+                raise_exception=True,
+            )
+        with self.assertRaises(PermissionDenied):
+            self.dispatch_view(
+                invalid_req_2,
+                permission_required=permissions,
+                object_level_permissions=True,
+                raise_exception=True,
+            )
+
 
 @pytest.mark.django_db
-class TestMultiplePermissionsRequiredMixin(
-    _TestAccessBasicsMixin, test.TestCase
-):
+class TestMultiplePermissionsRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
     """Scenarios around requiring multiple permissions"""
+
     view_class = MultiplePermissionsRequiredView
     view_url = "/multiple_permissions_required/"
 
@@ -534,10 +571,95 @@ class TestMultiplePermissionsRequiredMixin(
                 permissions=permissions,
             )
 
+    def test_all_object_level_permissions_key(self):
+        """
+        Tests that when a user has all the correct object level permissions, response is OK,
+        else forbidden.
+        """
+        # Arrange
+        article = ArticleFactory()
+        self.view_class = MultiplePermissionsRequiredView
+        self.view_url = f"/multiple_object_level_permissions_required/?pk={article.pk}"
+        auth_add_user = Permission.objects.get(codename="add_user")
+        tests_add_article = Permission.objects.get(codename="add_article")
+        permissions = {"all": ["auth.add_user", "tests.add_article"]}
+        valid_user = UserFactory(permissions=permissions["all"])
+        invalid_user = UserFactory(permissions=["auth.add_user"])
+        UserObjectPermissionsFactory(
+            user=valid_user, permission=auth_add_user, article_object=article
+        )
+        UserObjectPermissionsFactory(
+            user=valid_user, permission=tests_add_article, article_object=article
+        )
+        # Act
+        valid_req = self.build_request(path=self.view_url, user=valid_user)
+        valid_resp = self.dispatch_view(
+            valid_req, permissions=permissions, object_level_permissions=True
+        )
+        invalid_req = self.build_request(path=self.view_url, user=invalid_user)
+        # Arrange
+        self.assertEqual(valid_resp.status_code, 200)
+        with self.assertRaises(PermissionDenied):
+            self.dispatch_view(
+                invalid_req,
+                permissions=permissions,
+                object_level_permissions=True,
+                raise_exception=True,
+            )
+
+    def test_any_object_level_permissions_key(self):
+        """
+        Tests that when a user has any the correct object level permissions, response is OK,
+        else forbidden.
+        """
+        # Arrange
+        article = ArticleFactory()
+        self.view_url = f"/multiple_object_level_permissions_required/?pk={article.pk}"
+        self.view_class = MultiplePermissionsRequiredView
+        auth_add_user = Permission.objects.get(codename="add_user")
+        tests_add_article = Permission.objects.get(codename="add_article")
+        permissions = {"any": ["auth.add_user", "tests.add_article"]}
+        user = UserFactory(permissions=[permissions["any"][0]])
+        user_1 = UserFactory()
+        user_2 = UserFactory(permissions=permissions["any"])
+        UserObjectPermissionsFactory(
+            user=user, permission=auth_add_user, article_object=article
+        )
+        UserObjectPermissionsFactory(
+            user=user, permission=tests_add_article, article_object=article
+        )
+        # Act
+        valid_req = self.build_request(path=self.view_url, user=user)
+        valid_resp = self.dispatch_view(
+            valid_req,
+            permissions=permissions,
+            object_level_permissions=True,
+            raise_exception=True,
+        )
+        invalid_req_1 = self.build_request(path=self.view_url, user=user_1)
+        invalid_req_2 = self.build_request(path=self.view_url, user=user_2)
+        # Assert
+        self.assertEqual(valid_resp.status_code, 200)
+        with self.assertRaises(PermissionDenied):
+            self.dispatch_view(
+                invalid_req_1,
+                permissions=permissions,
+                object_level_permissions=True,
+                raise_exception=True,
+            )
+        with self.assertRaises(PermissionDenied):
+            self.dispatch_view(
+                invalid_req_2,
+                permissions=permissions,
+                object_level_permissions=True,
+                raise_exception=True,
+            )
+
 
 @pytest.mark.django_db
 class TestSuperuserRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
     """Scenarios requiring a superuser"""
+
     view_class = SuperuserRequiredView
     view_url = "/superuser_required/"
 
@@ -553,6 +675,7 @@ class TestSuperuserRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
 @pytest.mark.django_db
 class TestStaffuserRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
     """Scenarios requiring a staff user"""
+
     view_class = StaffuserRequiredView
     view_url = "/staffuser_required/"
 
@@ -655,6 +778,7 @@ class TestGroupRequiredMixin(_TestAccessBasicsMixin, test.TestCase):
 @pytest.mark.django_db
 class TestUserPassesTestMixin(_TestAccessBasicsMixin, test.TestCase):
     """Scenarios requiring a user to pass a test"""
+
     view_class = UserPassesTestView
     view_url = "/user_passes_test/"
     view_not_implemented_class = UserPassesTestNotImplementedView
@@ -708,6 +832,7 @@ class TestUserPassesTestMixin(_TestAccessBasicsMixin, test.TestCase):
 @pytest.mark.django_db
 class TestSSLRequiredMixin(test.TestCase):
     """Scenarios around requiring SSL"""
+
     view_class = SSLRequiredView
     view_url = "/sslrequired/"
 
@@ -744,7 +869,7 @@ class TestSSLRequiredMixin(test.TestCase):
 
 @pytest.mark.django_db
 class TestRecentLoginRequiredMixin(test.TestCase):
-    """ Scenarios requiring a recent login"""
+    """Scenarios requiring a recent login"""
 
     view_class = RecentLoginRequiredView
     recent_view_url = "/recent_login/"
@@ -769,7 +894,10 @@ class TestRecentLoginRequiredMixin(test.TestCase):
         user = UserFactory(last_login=last_login)
         self.client.login(username=user.username, password="asdf1234")
         resp = self.client.get(self.outdated_view_url)
-        assert resp.status_code == 302
+        assert resp.status_code in [
+            302,
+            405,
+        ]  # 302 is for Django < 5, while 405 is for Django >= 5
 
     def test_not_logged_in(self):
         """Anonymous requests should be handled appropriately"""
